# Copyright 2017 Shortest Track Company. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements BigQuery related data parsing helpers."""

import datetime


class Parser(object):
    """A set of helper functions to parse data in BigQuery responses."""

    def __init__(self):
        pass

    @staticmethod
    def parse_row(schema, data):
        """Parses a row from query results into an equivalent object.

        Args:
          schema: the array of fields defining the schema of the data.
          data: the JSON row from a query result.
        Returns:
          The parsed row object.
        """

        def parse_value(data_type, value):
            """Parses a value returned from a BigQuery response.

            Args:
              data_type: the type of the value as specified by the schema.
              value: the raw value to return (before casting to data_type).

            Returns:
              The value cast to the data_type.
            """
            if value is not None:
                if value == 'null':
                    value = None
                elif data_type == 'INTEGER':
                    value = int(value)
                elif data_type == 'FLOAT':
                    value = float(value)
                elif data_type == 'TIMESTAMP':
                    value = datetime.datetime.utcfromtimestamp(float(value))
                elif data_type == 'BOOLEAN':
                    value = value == 'true'
                elif (type(value) != str):
                    # TODO(gram): Handle nested JSON records
                    value = str(value)
            return value

        row = {}
        if data is None:
            return row

        for i, (field, schema_field) in enumerate(zip(data['f'], schema)):
            if field:
                val = field['v']
                name = schema_field['name']
                data_type = schema_field['type']
                repeated = True if 'mode' in schema_field and schema_field['mode'] == 'REPEATED' else False

                if repeated and val is None:
                    row[name] = []
                elif data_type == 'RECORD':
                    sub_schema = schema_field['fields']
                    if repeated:
                        row[name] = [Parser.parse_row(sub_schema, v['v']) for v in val]
                    else:
                        row[name] = Parser.parse_row(sub_schema, val)
                elif repeated:
                    row[name] = [parse_value(data_type, v['v']) for v in val]
                else:
                    row[name] = parse_value(data_type, val)
            else:
                name = schema_field['name']
                row[name] = []

        return row

    @staticmethod
    def parse_timestamp(value):
        """Parses a timestamp.

        Args:
          value: the number of milliseconds since epoch.
        """
        return datetime.datetime.utcfromtimestamp(float(value) / 1000.0)



# Copyright 2017 Shortest Track Company. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements BigQuery HTTP API wrapper."""

import time

from st_library.utils.helpers.store import Store
from st_library.utils.api_client import ApiClient


class StructuredDataService(object):
    """A helper class to issue BigQuery HTTP requests."""

    def sec_tables_get_by_sql(self, sql_statement, start_index=None, max_results=None, page_token=None):
        """Posts a SQL request to retrieve query_job_id

        Args:
            sql_statement: .
        Returns:
            A parsed result object.
        Raises:
            Exception if there is an error performing the operation.
        """
        url = ApiClient.res.sec_get_query_id(Store.config_id)

        body_sql = {'allowLargeResults': False,
                    'query': sql_statement}

        params = {}
        if start_index:
            params['startIndex'] = start_index
        if max_results:
            params['maxResults'] = max_results
        if page_token is not None:
            params['pageToken'] = page_token

        return ApiClient.post(url, json=body_sql, params=params)

    def tables_get(self, dataset_id, matrices_id):
        """Issues a request to retrieve information about a table.

        Args:
          dataset_id: .
          matrices_id: .
        Returns:
          A parsed result object.
        Raises:
          Exception if there is an error performing the operation.
        """
        return ApiClient.get(ApiClient.res.matrices_get(dataset_id, matrices_id))

    def tables_get_by_sql(self, query_job_id):
        """Issues a request to retrieve information about a table.

        Args:
          dataset_id: .
          matrices_id: .
        Returns:
          A parsed result object.
        Raises:
          Exception if there is an error performing the operation.
        """
        return ApiClient.get(ApiClient.res.matrices_get_by_sql(query_job_id))

    def tables_get_by_sql_query(self, query_job_id):
        """Issues a request to retrieve information about a table.

        Args:
          dataset_id: .
          matrices_id: .
        Returns:
          A parsed result object.
        Raises:
          Exception if there is an error performing the operation.
        """
        return ApiClient.post(ApiClient.res.sec_get_query_id(query_job_id))

    def tables_get_parameters(self):
        """Issues a request to retrieve the parameter of the script execution configuration.

        Args:
        Returns:
          A parsed result object.
        Raises:
          Exception if there is an error performing the operation.
        """
        return ApiClient.get(ApiClient.res.sec_get_detail(Store.config_id))

    def tabledata_list(self, config_related, datasetsid, table_name, start_index=None, max_results=None, page_token=None):
        """ Retrieves the contents of a table.

        Args:
          table_name: the name of the table as a tuple of components.
          start_index: the index of the row at which to start retrieval.
          max_results: an optional maximum number of rows to retrieve.
          page_token: an optional token to continue the retrieval.
        Returns:
          A parsed result object.
        Raises:
          Exception if there is an error performing the operation.
        """
        if config_related:
            url = ApiClient.res.sec_matrices_get(Store.config_id, datasetsid, table_name)
        else:
            url = ApiClient.res.matrices_get(datasetsid, table_name)
        params = {}
        if start_index:
            params['startIndex'] = start_index
        if max_results:
            params['maxResults'] = max_results
        if page_token is not None:
            params['pageToken'] = page_token

        return ApiClient.get(url, params=params)

    def tabledata_list_by_sql(self, query_job_id, start_index=None, max_results=None, page_token=None):
        """ Retrieves the contents of a table.

        Args:
          table_name: the name of the table as a tuple of components.
          start_index: the index of the row at which to start retrieval.
          max_results: an optional maximum number of rows to retrieve.
          page_token: an optional token to continue the retrieval.
        Returns:
          A parsed result object.
        Raises:
          Exception if there is an error performing the operation.
        """

        url = ApiClient.res.sec_matrices_get_by_sql(query_job_id)

        params = {}
        if start_index:
            params['startIndex'] = start_index
        if max_results:
            params['maxResults'] = max_results
        if page_token is not None:
            params['pageToken'] = page_token

        return ApiClient.get(url, params=params)

    def tabledata_post(self, datasetsid, table_name, file_name):
        """ Insert the contents of a table.

        Args:
          file_name: the name of the table as a tuple of components.
        Returns:
          A result object.
        Raises:
          Exception if there is an error performing the operation.
        """

        filepath1 = r"/home/st/workspace/st-dataprovider-python/datalab/structured_data/metadata.json"
        filepath2 = file_name
        boundary = '----------%s' % hex(int(time.time() * 1000))
        data = []
        data.append('--%s' % boundary)
        fr1 = open(filepath1, 'rb')
        data.append('Content-Disposition: form-data; name="%s"; filename="metadata.json"' % 'metadata')
        data.append('Content-Type: %s\r\n' % 'application/json')
        data.append(fr1.read())
        fr1.close()
        data.append('--%s' % boundary)
        fr2 = open(filepath2, 'rb')
        data.append('Content-Disposition: form-data; name="%s"; filename="data.csv"' % 'file')
        data.append('Content-Type: %s\r\n' % 'application/vnd.ms-excel')
        data.append(fr2.read())
        fr2.close()
        data.append('--%s--\r\n' % boundary)

        http_body = '\r\n'.join(data)

        headers = {}
        headers['Content-Type'] = 'multipart/form-data; boundary=%s' % boundary

        return ApiClient.post(
            ApiClient.res.matrices_upload(datasetsid, table_name), data=http_body, extra_headers=headers
        )

    def insert_sec_data(self, datasetsid, table_name, json_data):
        """ Insert streams data into matrix.

        Args:
          file_name: the name of the table as a tuple of components.
        Returns:
          A result object.
        Raises:
          Exception if there is an error performing the operation.
        """

        return ApiClient.post(ApiClient.res.sec_matrices_insert(
            Store.config_id, datasetsid, table_name), json=json_data
        )

    def insert_batch_sec_data(self, datasetsid, table_name, file_path, file_name):
        """ Insert streams data into matrix.

        Args:
          file_name: the name of the table as a tuple of components.
        Returns:
          A result object.
        Raises:
          Exception if there is an error performing the operation.
        """

        metadata_file = '/home/st/workspace/dataprovider-py/samples/data/structured_data/metadata_sec.json'
        data_file = file_path + file_name
        boundary = '----------%s' % hex(int(time.time() * 1000))
        data = []
        data.append('--%s' % boundary)
        fr1 = open(metadata_file, 'rb')
        data.append('Content-Disposition: form-data; name="%s"; filename="metadata_sec.json"' % 'metadata')
        data.append('Content-Type: %s\r\n' % 'application/json')
        data.append(fr1.read())
        fr1.close()
        data.append('--%s' % boundary)
        fr2 = open(data_file, 'rb')
        data.append('Content-Disposition: form-data; name="%s"; filename="data.csv"' % 'data')
        data.append('Content-Type: %s\r\n' % 'application/vnd.ms-excel')
        data.append(fr2.read())
        fr2.close()
        data.append('--%s--\r\n' % boundary)

        http_body = '\r\n'.join(data)

        headers = {}
        headers['Content-Type'] = 'multipart/form-data; boundary=%s' % boundary

        return ApiClient.post(ApiClient.res.sec_matrices_batch_insert(
            Store.config_id, datasetsid, table_name), data=http_body, extra_headers=headers
        )



# Copyright 2017 Shortest Track Company. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements Table, and related Table BigQuery APIs."""
import pandas

from st_library.core.dataprovider.structured_data import parser
from st_library.core.dataprovider.structured_data import schema
from st_library.utils.api_client.services.structured_data import StructuredDataService
from st_library import exceptions
from time import sleep


# import of Query is at end of module as we have a circular dependency of
# Query.execute().results -> Table and Table.sample() -> Query

class Table(object):
    """
    Represents a Table object referencing a BigQuery table.

    Parameters
    ----------
    matrix_id : str
    dataset_id : str
    name : str
        The name of the table
    context
    config_related : bool
        This parameter determines whether the table is related to
        Script Execution Configuration or not

    Attributes
    ----------
    schema : :class:`~st_library.core.dataprovider.structured_data.schema.Schema`
        A Schema object containing a list of schema fields and associated metadata.

    """

    # Allowed characters in a BigQuery table column name
    _VALID_COLUMN_NAME_CHARACTERS = '_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'

    # When fetching table contents, the max number of rows to fetch per HTTP request
    _DEFAULT_PAGE_SIZE = 1024

    # Milliseconds per week
    _MSEC_PER_WEEK = 7 * 24 * 3600 * 1000

    def __init__(self, matrix_id=None, dataset_id=None, name=None, context=None, config_related=False):
        self._context = context
        self._service = StructuredDataService()
        self._name_parts = name
        self._datasets_id = dataset_id
        self._matrices_id = matrix_id
        self._info = None
        self._cached_page = None
        self._cached_page_index = 0
        self._schema = None
        self._config_related = config_related

    def _load_info(self):
        """
        Loads metadata about this table.
        """
        if self._info is None:
            try:
                self._info = self._service.tables_get(self._datasets_id, self._matrices_id)
            except Exception as e:
                raise e

    def _load_info_by_sql(self):
        """
        Loads metadata about this table.
        """
        if self._info is None:
            try:
                self._info = self._service.tables_get_by_sql("54a856b9-b1c6-4651-84ff-1f9428965a47")
            except Exception as e:
                raise e

    def _load_info_by_sql_query(self, sql_query):
        """
        Loads metadata about this table.
        """
        if self._info is None:
            try:
                self._info = self._service.sec_tables_get_by_sql(sql_query, start_index=0, max_results=2)
            except Exception as e:
                raise e

    def _load_info_by_response(self, response):
        """
        Loads metadata about this table.
        """
        if self._info is None:
            self._info = response

    def _get_row_fetcher(self, start_row=0, max_rows=None, page_size=_DEFAULT_PAGE_SIZE):
        """
        Get a function that can retrieve a page of rows.

        Parameters
        ----------
            start_row: int
                the row to start fetching from; default 0.

            max_rows: int
                the maximum number of rows to fetch (across all calls, not per-call). Default
                is None which means no limit.

            page_size: int
                the maximum number of results to fetch per page; default 1024.

        Returns
        -------
            `callable` A function that can be called repeatedly with a page token and running count, and that
            will return an array of rows and a next page token; when the returned page token is None
            the fetch is complete.

        Notes
        -----
            The function returned is a closure so that it can have a signature suitable for use
            by Iterator.

        """
        if not start_row:
            start_row = 0
        elif start_row < 0:  # We are measuring from the table end
            if self.length >= 0:
                start_row += self.length
            else:
                raise Exception('Cannot use negative indices for table of unknown length')

        schema = self.schema._bq_schema
        name_parts = self._name_parts

        def _retrieve_rows(page_token, count):

            page_rows = []
            if max_rows and count >= max_rows:
                page_token = None
            else:
                if max_rows and page_size > (max_rows - count):
                    max_results = max_rows - count
                else:
                    max_results = page_size

                try:
                    if page_token:
                        response = self._service.tabledata_list(self._config_related, self._datasets_id,
                                                                name_parts, page_token=page_token,
                                                                max_results=max_results)
                    else:
                        response = self._service.tabledata_list(self._config_related, self._datasets_id,
                                                                name_parts, start_index=start_row,
                                                                max_results=max_results)
                except Exception as e:
                    raise e
                page_token = response['pageToken'] if 'pageToken' in response else None
                if 'rows' in response:
                    page_rows = response['rows']

            rows = []
            for row_dict in page_rows:
                rows.append(parser.Parser.parse_row(schema, row_dict))

            return rows, page_token

        return _retrieve_rows

    def _get_row_fetcher_by_sql(self, start_row=0, max_rows=None, page_size=_DEFAULT_PAGE_SIZE, job_query_id=None):
        """
        Get a function that can retrieve a page of rows.

        Parameters
        ----------
            start_row: int
                the row to start fetching from; default 0.

            max_rows: int
                the maximum number of rows to fetch (across all calls, not per-call). Default
                is None which means no limit.

            page_size: int
                the maximum number of results to fetch per page; default 1024.

            job_query_id: str
                the string containing id of a job query to BQ


        Returns
        -------
            `callable` A function that can be called repeatedly with a page token and running count, and that
            will return an array of rows and a next page token; when the returned page token is None
            the fetch is complete.

        Notes
        -----
            The function returned is a closure so that it can have a signature suitable for use
            by Iterator.

        """
        if not start_row:
            start_row = 0
        elif start_row < 0:  # We are measuring from the table end
            if self.length >= 0:
                start_row += self.length
            else:
                raise Exception('Cannot use negative indices for table of unknown length')

        schema = self.schema_by_sql._bq_schema
        name_parts = self._name_parts

        def _retrieve_rows(page_token, count):
            page_size = 5000
            page_rows = []
            if max_rows and count >= max_rows:
                page_token = None
            else:
                if max_rows and page_size > (max_rows - count):
                    max_results = max_rows - count
                else:
                    max_results = page_size

                try:
                    if page_token:
                        response = self._service.tabledata_list_by_sql(job_query_id, page_token=page_token,
                                                                max_results=max_results)
                    else:
                        response = self._service.tabledata_list_by_sql(job_query_id, start_index=start_row,
                                                                max_results=max_results)
                except Exception as e:
                    raise e
                page_token = response['pageToken'] if 'pageToken' in response else None
                if 'rows' in response:
                    page_rows = response['rows']

            rows = []
            for row_dict in page_rows:
                rows.append(parser.Parser.parse_row(schema, row_dict))

            return rows, page_token

        return _retrieve_rows

    def _get_row_fetcher_by_sql_query(self, start_row=0, max_rows=None, page_size=_DEFAULT_PAGE_SIZE, sql_query=None):
        """
        Get a function that can retrieve a page of rows.

        Parameters
        ----------
            start_row: int
                the row to start fetching from; default 0.

            max_rows: int
                the maximum number of rows to fetch (across all calls, not per-call). Default
                is None which means no limit.

            page_size: int
                the maximum number of results to fetch per page; default 1024.

            job_query_id: str
                the string containing id of a job query to BQ


        Returns
        -------
            `callable` A function that can be called repeatedly with a page token and running count, and that
            will return an array of rows and a next page token; when the returned page token is None
            the fetch is complete.

        Notes
        -----
            The function returned is a closure so that it can have a signature suitable for use
            by Iterator.

        """
        if not start_row:
            start_row = 0
        elif start_row < 0:  # We are measuring from the table end
            if self.length >= 0:
                start_row += self.length
            else:
                raise Exception('Cannot use negative indices for table of unknown length')

        name_parts = self._name_parts

        def _retrieve_rows(page_token, count):
            page_size = 5000
            page_rows = []

            # print('max rows:')
            # print(max_rows)
            # print('count:')
            # print(count)
            # print('page token:')
            # print(page_token)

            if max_rows and count >= max_rows:
                page_token = None
                return [], page_token
            else:
                if max_rows and page_size > (max_rows - count):
                    max_results = max_rows - count
                else:
                    max_results = page_size
                try:
                    if page_token:
                        response = self._service.sec_tables_get_by_sql(sql_query, page_token=page_token,
                                                                       max_results=max_results)
                        if not response['jobComplete']:
                            if response.get('jobReference'):
                                job_query_id = response['jobReference']['jobId']
                            else:
                                job_query_id = response['id']

                            while True:
                                response = self._service.tabledata_list_by_sql(job_query_id, page_token=page_token,
                                                                               max_results=max_results)
                                if response['jobComplete']:
                                    break
                                sleep(1.5)

                    else:
                        response = self._service.sec_tables_get_by_sql(sql_query, start_index=start_row,
                                                                       max_results=max_results)
                        if not response['jobComplete']:
                            if response.get('jobReference'):
                                job_query_id = response['jobReference']['jobId']
                            else:
                                job_query_id = response['id']

                            while True:
                                response = self._service.tabledata_list_by_sql(job_query_id, start_index=start_row,
                                                                               max_results=max_results)
                                if response['jobComplete']:
                                    break
                                sleep(1.5)

                except Exception as e:
                    raise e
                page_token = response['pageToken'] if 'pageToken' in response else None
                if 'rows' in response:
                    page_rows = response['rows']

            rows = []

            bq_schema = self.schema_by_response(response).bq_schema

            for row_dict in page_rows:
                rows.append(parser.Parser.parse_row(bq_schema, row_dict))

            return rows, page_token

        return _retrieve_rows


    def to_dataframe(self, start_row=0, max_rows=200):
        """
        Exports the table to a Pandas dataframe.

        Parameters
        ----------
            start_row: int
                the row of the table at which to start the export (default 0)

            max_rows: int
                an upper limit on the number of rows to export (default None)

        Returns
        -------
            :class:`pandas.DataFrame` containing the table data.

        """
        fetcher = self._get_row_fetcher(start_row=start_row, max_rows=max_rows)
        count = 0
        page_token = None
        df = None
        while True:
            page_rows, page_token = fetcher(page_token, count)
            if len(page_rows):
                count += len(page_rows)
                if df is None:
                    df = pandas.DataFrame.from_records(page_rows)
                else:
                    df = df.append(page_rows, ignore_index=True)
            if not page_token:
                break

        # Need to reorder the dataframe to preserve column ordering
        ordered_fields = [field.name for field in self.schema]
        return df[ordered_fields] if df is not None else pandas.DataFrame()

    def to_dataframe_by_sql(self, start_row=0, max_rows=None, sql_query = None):
        """
        Exports the table to a Pandas dataframe.

        Parameters
        ----------
            start_row: int
                the row of the table at which to start the export (default 0)

            max_rows: int
                an upper limit on the number of rows to export (default None)

        Returns
        -------
            :class:`pandas.DataFrame` containing the table data.

        """
        fetcher = self._get_row_fetcher_by_sql_query(sql_query=sql_query, start_row=start_row, max_rows=max_rows)
        count = 0
        page_token = None
        df = None
        while True:
            page_rows, page_token = fetcher(page_token, count)
            if len(page_rows):
                count += len(page_rows)
                if df is None:
                    df = pandas.DataFrame.from_records(page_rows)
                else:
                    df = df.append(page_rows, ignore_index=True)
            if not page_token:
                break

        # Need to reorder the dataframe to preserve column ordering
        ordered_fields = [field.name for field in self.schema]
        return df[ordered_fields] if df is not None else pandas.DataFrame()

    def get_query_id_token(self, sql_query):
        """
        Retrieves query_job_id from a SQL query

        Parameters
        ----------
            sql_statement: str
                SQL query for a BigQuery

        Returns
        -------
            query_job_id

        """
        response = self._service.sec_tables_get_by_sql(sql_query=sql_query)
        print(response['id'])

        return response['id']
        #return response['jobReference']['jobId']


    @property
    def schema(self):
        if not self._schema:
            try:
                self._load_info()
                self._schema = schema.Schema(self._info['schema']['fields'])

            except KeyError:
                raise exceptions.InternalError('Unexpected table response: missing schema')
        return self._schema

    @property
    def schema_by_sql(self):
        if not self._schema:
            try:
                self._load_info_by_sql()
                self._schema = schema.Schema(self._info['schema']['fields'])

            except KeyError:
                raise exceptions.InternalError('Unexpected table response: missing schema')
        return self._schema

    def schema_by_sql_query(self, sql_query):
        if not self._schema:
            try:
                self._load_info_by_sql_query(sql_query)
                self._schema = schema.Schema(self._info['schema']['fields'])

            except KeyError:
                raise exceptions.InternalError('Unexpected table response: missing schema. Response content: {}'.
                                               format(self._info))
        return self._schema

    def schema_by_response(self, response):
        if not self._schema:
            try:
                self._load_info_by_response(response)
                self._schema = schema.Schema(self._info['schema']['fields'])

            except KeyError:
                raise exceptions.InternalError('Unexpected table response: missing schema. Response content: {}'.
                                               format(self._info))
        return self._schema

    def upload_data(self, filename=None):
        """
        Upload csv data file to table.

        Parameters
        ----------
        filename: str
            the name of the csv file

        """
        response = self._service.tabledata_post(self._datasets_id, self._name_parts, filename)

    def insert_data(self, json_data=None):
        """
        Insert streams data into matrix.

        Parameters
        ----------
        json_data: object
            the records of the matrix

        """
        response = self._service.insert_data(self._datasets_id, self._name_parts, json_data)

    def insert_sec_data(self, json_data=None):
        """
        Insert streams data into matrix.

        Parameters
        ----------
        json_data: object
            the records of the matrix

        """
        response = self._service.insert_sec_data(self._datasets_id, self._name_parts, json_data)

    def insert_batch_sec_data(self, file_path=None, file_name=None):
        """
        Insert streams data into matrix.

        Parameters
        ----------
        file_path: str
            the path to the csv file

        file_name: str
            the name of the csv file

        """
        response = self._service.insert_batch_sec_data(self._datasets_id, self._name_parts, file_path, file_name)



from st_library.core.dataprovider.unstructured_data.item import Item


class UnstructuredData(object):
    """
    A Category for work with unstructured data

    Attributes
    ----------
    Item : object
        The :class:`~st_library.core.dataprovider.unstructured_data.item.Item` class itself

    """

    def __init__(self):
        self.Item = Item

    def download_file(self, dataset_id, filename):
        """
        Download a file to disk.

        Parameters
        ----------
        dataset_id : str
        filename : str

        Returns
        -------
        str
            Local file path

        """
        item = self.Item(dataset_id=dataset_id)
        return item.download_file(filename)

    def upload_file(self, dataset_id, filename, filepath):
        """
        Upload a file.

        Parameters
        ----------
        dataset_id : str
        filename : str
        filepath : str

        """
        item = self.Item(dataset_id=dataset_id)
        item.upload_file(filename, filepath)

    def delete_file(self, dataset_id, filename):
        """
        Delete a file.

        Parameters
        ----------
        dataset_id : str
        filename : str

        """
        item = self.Item(dataset_id=dataset_id)
        item.delete_file(filename)



from st_library.utils.api_client import ApiClient
from st_library.utils.generics.base_object import BaseModel
from st_library.utils.helpers.store import Store


class ScriptExecutionConfiguration(BaseModel):
    def get_initial_data(self):
        return ApiClient.get(ApiClient.res.sec_get_detail(Store.config_id))

    def get_parameter_value(self, parameter_id):
        for parameter in self['parameters']:
            if parameter['id'] == parameter_id:
                return parameter['value']

        raise ValueError('Parameter {} does not exist'.format(parameter_id))



from st_library.utils.api_client import resource
from st_library.utils.api_client import request


class ApiClient(object):
    def __init__(self):
        raise TypeError('This class should not be initialized')

    res = resource

    @staticmethod
    def get(*args, **kwargs):
        return request.get(*args, **kwargs)

    @staticmethod
    def post(*args, **kwargs):
        return request.post(*args, **kwargs)

    @staticmethod
    def delete(*args, **kwargs):
        return request.delete(*args, **kwargs)



import collections


class BaseModel(collections.Mapping):
    def get_initial_data(self):
        raise NotImplementedError

    def __init__(self):
        self._data = None
        self._data = self.get_initial_data()

    def __getitem__(self, key):
        return self._data[key]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)



import abc

import six

from . import resources


@six.add_metaclass(abc.ABCMeta)
class PublicKeyMixin(object):
    @abc.abstractmethod
    def _get_cache_public_key(self):
        """
        Return public key from cache

        """

    @abc.abstractmethod
    def _set_cache_public_key(self, public_key):
        """
        Set public key to cache

        """

    @abc.abstractmethod
    def _reset_cache_public_key(self):
        """
        Reset public key to cache

        """

    def get_public_key(self):
        public_key = self._get_cache_public_key()
        if public_key:
            return public_key

        public_key_bytes = self.request(self.GET, 'token_publickey', base=resources.JWTKEY_ENDPOINT,
                                        extra_headers={'Authorization': None}, raw_content=True)
        public_key = public_key_bytes.decode('utf-8')
        self._set_cache_public_key(public_key)

        return public_key



from urlobject import URLObject
from st_library.utils.helpers.store import Store


ROOT = URLObject(Store.host)

METADATA_ENDPOINT = ROOT.add_path('api/metadata/')
EXECUTION_METADATA_ENDPOINT = ROOT.add_path('api/execution-metadata/')
LOGGING_ENDPOINT = ROOT.add_path('api/logging/')
DATA_ENDPOINT = ROOT.add_path('api/data')

GETTOKEN_ENDPOINT = ROOT.add_path('api/uaa/api-keys/')


PERFORMANCES_LIST_GET_PATH = EXECUTION_METADATA_ENDPOINT.add_path('v2/performances/')

def sec_get_token(uuid):
    return GETTOKEN_ENDPOINT.add_path('{uuid}'.format(uuid=uuid))


def sec_get_detail(uuid):
    return METADATA_ENDPOINT.add_path('script-execution-configurations/{uuid}'.format(uuid=uuid))


def sec_get_query_id(uuid):
    return DATA_ENDPOINT.add_path('script-execution-configurations/{uuid}/queries'.format(uuid=uuid))


def performance_get_detail(uuid):
    return EXECUTION_METADATA_ENDPOINT.add_path('v2/performances/{uuid}/'.format(uuid=uuid))


def performance_output_parameter_post(parameter_id, performance_id):
    return EXECUTION_METADATA_ENDPOINT.\
        add_path('v2/performances/{performance_id}/output-parameters/{parameter_id}/value/'.
                 format(performance_id=performance_id, parameter_id=parameter_id))


def logging_post(performance_id):
    return LOGGING_ENDPOINT.add_path('performances/{performance_id}/'.format(performance_id=performance_id))


def matrices_upload(dataset_id, matrices_id):
    return DATA_ENDPOINT.add_path('script-execution-configurations/{sec_id}/'
                                  'datasets/{dataset_id}/matrices/{matrices_id}/upload'.
                                  format(sec_id=Store.config_id, dataset_id=dataset_id, matrices_id=matrices_id))


def matrices_get(dataset_id, matrices_id):
    return DATA_ENDPOINT.add_path('script-execution-configurations/{sec_id}/'
                                  'datasets/{dataset_id}/matrices/{matrices_id}/data'.
                                  format(sec_id=Store.config_id, dataset_id=dataset_id, matrices_id=matrices_id))

def matrices_get_by_sql(query_job_id):
    return DATA_ENDPOINT.add_path('queries/{query_job_id}'.
                                  format(query_job_id=query_job_id))


def sec_matrices_get(sec_id, dataset_id, matrices_id):
    return DATA_ENDPOINT.add_path('script-execution-configurations/{sec_id}/datasets/'
                                  '{dataset_id}/matrices/{matrices_id}/data'.
                                  format(sec_id=sec_id, dataset_id=dataset_id, matrices_id=matrices_id))


def sec_matrices_get_by_sql(query_job_id):
    return DATA_ENDPOINT.add_path('script-execution-configurations/{sec_id}/queries/{query_job_id}'.
                                  format(sec_id=Store.config_id, query_job_id=query_job_id))


def sec_matrices_insert(sec_id, dataset_id, matrices_id):
    return DATA_ENDPOINT.add_path('script-execution-configurations/{sec_id}/datasets/'
                                  '{dataset_id}/matrices/{matrices_id}/insert'.
                                  format(sec_id=sec_id, dataset_id=dataset_id, matrices_id=matrices_id))


def sec_matrices_batch_insert(sec_id, dataset_id, matrices_id):
    return DATA_ENDPOINT.add_path('script-execution-configurations/{sec_id}/datasets/'
                                  '{dataset_id}/matrices/{matrices_id}/upload'.
                                  format(sec_id=sec_id, dataset_id=dataset_id, matrices_id=matrices_id))


def object_download(repo_id, file_id):
    return DATA_ENDPOINT.add_path('blob-repositories/{repo_id}/download'.
                                  format(repo_id=repo_id)).add_query_param('name', file_id)


def object(repo_id, file_id):
    return DATA_ENDPOINT.add_path('blob-repositories/{repo_id}'.
                                  format(repo_id=repo_id)).add_query_param('name', file_id)



# Copyright 2017 Shortest Track Company. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

# To publish to PyPi use: python setup.py bdist_wheel upload -r pypi

from setuptools import setup, find_packages

version = '1.0.20'

setup(
    name='st-dataprovider',
    version=version,
    packages=find_packages(),
    description='shortest track dataprovider',
    author='Fugui Xing',
    author_email='admin@shortesttrack.com',
    url='https://github.com/shortesttrack/dataprovider-py',
    download_url='https://github.com/shortesttrack/dataprovider-py/archive/v1.0.20',
    keywords=[
        'Shortest Track',
        'st_library',
        'dataprovider'
    ],
    license="Apache Software License",
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    long_description="""\
Support package for Shortest Track Library. This provides Python APIs
for accessing Shortest Track Platform services such as structured data storage.
  """,
    install_requires=[
        'requests-toolbelt==0.8.0',
        'pandas==0.19.1',
        'urlobject==2.4.3',
        'six==1.11.0',
        'python-dateutil==2.7.0',
        'PyJWT==1.6.1',
        'cryptography==2.2.1',
        'mock==2.0.0',
        'pytz==2018.3',
        'psycopg2-binary==2.7.4',
        'redis==2.10.6',
        'backoff==1.5.0'
    ]
)



# Copyright 2017 Shortest Track Company. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.

"""Implements BigQuery HTTP API wrapper."""

import os

from requests_toolbelt.multipart.encoder import MultipartEncoder

from st_library.utils.api_client import ApiClient


class UnstructuredDataService(object):
    """A helper class to issue BigQuery HTTP requests."""

    def __init__(self):
        """Initializes the BigQuery helper with context information.

        """

    def tables_get(self, table_name, matrices_id):
        """Issues a request to retrieve information about a table.

        Args:
          table_name: a tuple representing the full name of the table.
        Returns:
          A parsed result object.
        Raises:
          Exception if there is an error performing the operation.
        """

        return ApiClient.get(ApiClient.res.object(table_name, matrices_id))

    def download_object(self, datasetsid, file_name):
        """ Download an object.

        Args:
          datasetsid: the id of the dataset.
          file_name: the name of the dataset files.
        Returns:
          A url for downloading object.
        Raises:
          Exception if there is an error performing the operation.
        """

        return ApiClient.get(ApiClient.res.object_download(datasetsid, file_name), raw_response=True)

    def upload_object(self, datasetsid, file_name, file_path):
        """ Upload file to gcs.

        Args:
          file_name: the name of the file.
          file_path: the path of the file.
        Returns:
          A result object.
        Raises:
          Exception if there is an error performing the operation.
        """

        path = os.path.join(file_path, file_name)

        with open(path, 'rb') as f:
            multipart_data = MultipartEncoder(
                fields={
                    # a file upload field
                    'file': (file_name, f, 'text/plain')
                }
            )
            headers = dict()
            headers['Content-Type'] = multipart_data.content_type

            return ApiClient.post(ApiClient.res.object(datasetsid, file_name), data=multipart_data,
                                  extra_headers=headers)

    def delete_object(self, datasetsid, file_name):
        """ Delete an object.

        Args:
          datasetsid: the id of the dataset.
          file_name: the name of the dataset files.
        Returns:
          A url for downloading object.
        Raises:
          Exception if there is an error performing the operation.
        """

        return ApiClient.delete(ApiClient.res.object(datasetsid, file_name))



import redis

from st_library.utils.generics.connectors import ConnectorContainer


class Redis(object):
    def __init__(self, db, host, port, password):
        self._db = db
        self._host = host
        self._port = port
        self._password = password
        self._redis = redis.StrictRedis(host=host, port=port, db=db, password=password)

    def __repr__(self):
        return '<Postgres db "{}">'.format(self._db)

    @property
    def db(self):
        return self.db

    @property
    def redis(self):
        return self._redis


class RedisContainer(ConnectorContainer):
    def _do_initialize_data(self):
        assert not len(self._list)

        params = self._fetch_param_dict([
            'redis_db', 'redis_host', 'redis_password', 'redis_port'
        ])

        list_of_servers_params = [params]

        for server in list_of_servers_params:
            obj = Redis(server['redis_db'], server['redis_host'],
                        int(server['redis_port']),
                        server['psql_password'])
            self._list.append(obj)



import backoff
import psycopg2
from psycopg2.extras import DictCursor, NamedTupleCursor

from st_library.utils.generics.connectors import ConnectorContainer


_disconnect_errors = (psycopg2.InterfaceError, psycopg2.OperationalError,)
_backoff = backoff.on_exception(backoff.expo, _disconnect_errors, max_time=30, max_tries=30)


class Postgres(object):
    DictCursor = DictCursor
    NamedTupleCursor = NamedTupleCursor

    def __init__(self, name, host, port, username, password):
        self._name = name
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._conn = None
        self._cursor = None
        self._cursor_type = None

    def __repr__(self):
        return '<Postgres db "{}">'.format(self._name)

    def set_cursor_type(self, cursor_type):
        self._cursor_type = cursor_type

    @property
    def name(self):
        return self._name

    @_backoff
    def _get_connection(self):
        if self._conn and not self._conn.closed:
            return self._conn
        db_connection = self._do_get_connection()
        self._conn = db_connection
        return self._conn

    def _do_get_connection(self):
        return psycopg2.connect(database=self._name,
                                user=self._username,
                                password=self._password,
                                host=self._host, port=self._port)

    @_backoff
    def execute(self, *args, **kwargs):
        if self._cursor is None or self._cursor.closed:
            self._cursor = self._get_connection().cursor(cursor_factory=self._cursor_type)
        return self._cursor.execute(*args, **kwargs)

    def fetchall(self):
        return self._cursor.fetchall()

    def fetchmany(self, *args, **kwargs):
        return self._cursor.fetchmany(*args, **kwargs)

    def fetchone(self):
        return self._cursor.fetchone()

    @_backoff
    def commit(self):
        self._get_connection().commit()

    @_backoff
    def cancel(self):
        self._get_connection().cancel()

    @_backoff
    def close(self):
        self._get_connection().close()

    @_backoff
    def rollback(self):
        self._get_connection().rollback()


class PostgresContainer(ConnectorContainer):
    def _do_initialize_data(self):
        assert not len(self._list)

        params = self._fetch_param_dict([
            'psql_host', 'psql_name', 'psql_username', 'psql_password', 'psql_port'
        ])

        list_of_servers_params = [params]

        for server in list_of_servers_params:
            obj = Postgres(server['psql_name'], server['psql_host'],
                           int(server['psql_port']), server['psql_username'],
                           server['psql_password'])
            self._list.append(obj)
            self._dict[obj.name] = obj



# Copyright 2017 Shortest Track Company. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing permissions and limitations under
# the License.
from st_library.core.dataprovider.structured_data.structured_data import StructuredData
from st_library.core.dataprovider.unstructured_data.unstructured_data import UnstructuredData
from st_library.core.logger.logger import Logger
from st_library.core.metadata.metadata import _METADATA
from st_library.utils.databases.databases import Databases
from st_library.utils.generics.singleton import Singleton
from st_library.utils.helpers.store import Store


__all__ = ('Library',)


class Library(Singleton):
    """
    The Library containing all the required functionality for work with Shortest Track Services.

    Attributes
    ----------
    struct_data : :class:`~st_library.core.dataprovider.structured_data.structured_data.StructuredData`
        A Category for work with structured data

    unstruct_data : :class:`~st_library.core.dataprovider.unstructured_data.unstructured_data.UnstructuredData`
        A Category for work with unstructured data

    logger : :class:`~st_library.core.logger.logger.Logger`
        A Category for log messages

    metadata : :class:`~st_library.core.metadata.metadata.Metadata`
        A Category for work with metadata

    config_id : str
        ID of the Script Execution Configuration for current Environment

    token : str
        The Token for current Environment

    performance_id : str
        ID of the Performance for current Environment

    Methods
    -------
    set_config_id(config_id='1686408f-2fc5-4657-b0d3-1c5c6b8e1ad7')
        Set the ID of Script Execution Configuration for current Environment

    set_token(token='1686408f-2fc5-4657-b0d3-1c5c6b8e1ad7')
        Set the Token for current Environment

    Examples
    --------
        >>> from st_library import Library
        >>> st_lib = Library()
        >>> st_lib.set_config_id('1686408f-2fc5-4657-b0d3-1c5c6b8e1ad7')
        >>> param = st_lib.metadata.get_parameter_value('foo')
        >>> table = st_lib.struct_data.Table('bar', 'baz', 'bar-baz')
        >>> table.upload_data('/foo/bar.csv')

    """

    def __init__(self):
        self.struct_data = StructuredData()
        self.unstruct_data = UnstructuredData()
        self.logger = Logger()
        self.metadata = _METADATA
        self.databases = Databases()

    @property
    def config_id(self):
        return Store.config_id

    def set_config_id(self, config_id):
        Store.config_id = config_id

    @property
    def token(self):
        return Store.token

    def set_token(self, token):
        Store.token = token

    @property
    def performance_id(self):
        return Store.performance_id



from st_library.core.dataprovider.structured_data.table import Table


class StructuredData(object):
    """
    A Category for work with structured data

    Attributes
    ----------
    Table : object
        The :class:`~st_library.core.dataprovider.structured_data.table.Table` class itself

    Examples
    --------
        >>> from st_library import Library
        >>> st_lib = Library()
        >>> table = st_lib.struct_data.Table('bar', 'baz', 'bar-baz')
        >>> table.upload_data('/foo/bar.csv')
        >>> sec_matrix = st_lib.struct_data.read_sec_matrix('foo', 'foo_dataset', 'foo_table_name')

    """

    def __init__(self):
        self.Table = Table

    def read_matrix(self, matrix_id, dataset_id, table_name):
        """
        Retrieve the table data from Global Space.

        Parameters
        ----------
        matrix_id : str
        dataset_id : str
        table_name : str

        Returns
        -------
        :class:`pandas.DataFrame`

        """
        tbl = self.Table(matrix_id, dataset_id, table_name)
        return tbl.to_dataframe()

    def read_sec_matrix(self, matrix_id, dataset_id, table_name):
        """
        Retrieve the table data from Script Execution Configuration Space.

        Parameters
        ----------
        matrix_id : str
        dataset_id : str
        table_name : str

        Returns
        -------
        :class:`pandas.DataFrame`

        """
        tbl = self.Table(matrix_id, dataset_id, table_name, config_related=True)
        return tbl.to_dataframe()

    def read_matrix_by_sql(self, sql_query, max_rows=None):
        """
        Retrieve the table data from Script Execution Configuration Space by sql statement.

        Parameters
        ----------
        sql : str

        Returns
        -------
        :class:`pandas.DataFrame`

        """
        # begin_date = datetime.date.today()-datetime.timedelta(duration)
        # regex_date = re.search(u'\d{4}\-\d{1,2}\-\d{1,2}', sql_query)
        # if regex_date:
        #     sql_date = datetime.datetime.strptime(regex_date[0], "%Y-%m-%d")
        #     if (datetime.datetime.today() - sql_date).days > 7:
        #         print('true')

        tbl = self.Table()
        return tbl.to_dataframe_by_sql(sql_query=sql_query, max_rows=max_rows)


from st_library.utils.databases.postgres import PostgresContainer
from st_library.utils.databases.redis import RedisContainer


class Databases(object):
    def __init__(self):
        self.psql_dbs = PostgresContainer()
        self.redis_dbs = RedisContainer()
